import os

from tempfile import mkstemp

import click
import six

import pyinfra

from pyinfra import local, logger
from pyinfra.api.exceptions import InventoryError
from pyinfra.api.util import get_file_io
from pyinfra.progress import progress_spinner

from .local import run_shell_command as run_local_shell_command


def make_names_data(image=None):
    if not image:
        raise InventoryError('No docker base image provided!')

    # Save the image as the hostname, no data, @docker group
    yield image, {}, ['@docker']


def connect(state, host, for_fact=None):
    if 'docker_container_id' in host.host_data:  # user can provide a docker_container_id
        return True

    with progress_spinner({'docker run'}):
        container_id = local.shell(
            'docker run -d {0} sleep 10000'.format(host.name),
            splitlines=True,
        )[-1]  # last line is the container ID

    host.host_data['docker_container_id'] = container_id
    return True


def disconnect(state, host):
    if not pyinfra.is_cli:
        return

    container_id = host.host_data['docker_container_id'][:12]

    with progress_spinner({'docker commit'}):
        image_id = local.shell(
            'docker commit {0}'.format(container_id),
            splitlines=True,
        )[-1][7:19]  # last line is the image ID, get sha256:[XXXXXXXXXX]...

    with progress_spinner({'docker rm'}):
        local.shell(
            'docker rm -f {0}'.format(container_id),
        )

    logger.info('{0}docker build complete, image ID: {1}'.format(
        host.print_prefix, click.style(image_id, bold=True),
    ))


def run_shell_command(
    state, host, command,
    timeout=None, print_output=False,
    **kwargs  # ignored (sudo/etc)
):
    container_id = host.host_data['docker_container_id']
    docker_command = 'docker exec {0} sh -c "{1}"'.format(container_id, command)

    return run_local_shell_command(
        state, host, docker_command,
        timeout=timeout,
        print_output=print_output,
    )


def put_file(
    state, host, filename_or_io, remote_filename,
    print_output=False,
    **kwargs  # ignored (sudo/etc)
):
    _, temp_filename = mkstemp()

    # Load our file or IO object and write it to the temporary file
    with get_file_io(filename_or_io) as file_io:
        with open(temp_filename, 'wb') as temp_f:
            data = file_io.read()

            if isinstance(data, six.text_type):
                data = data.encode()

            temp_f.write(data)

    docker_id = host.host_data['docker_container_id']
    docker_command = 'docker cp {0} {1}:{2}'.format(
        temp_filename,
        docker_id,
        remote_filename,
    )

    status, _, stderr = run_local_shell_command(
        state, host, docker_command,
        print_output=print_output,
    )

    if temp_filename:
        os.remove(temp_filename)

    if not status:
        raise IOError('\n'.join(stderr))

    if print_output:
        print('{0}file copied: {1}'.format(host.print_prefix, remote_filename))

    return status



# pyinfra
# File: pyinfra/facts/git.py
# Desc: local git repo facts

from __future__ import unicode_literals

from pyinfra.api.facts import FactBase


class GitBranch(FactBase):
    @staticmethod
    def command(repo):
        return 'cd {0} && git rev-parse --abbrev-ref HEAD'.format(repo)


class GitConfig(FactBase):
    @staticmethod
    def command(repo=None):
        if repo is None:
            return 'git config --global -l'

        return 'cd {0} && git config --local -l'.format(repo)

    @staticmethod
    def process(output):
        items = {}

        for line in output:
            key, value = line.split('=', 1)
            items[key] = value

        return items



# pyinfra
# File: pyinfra/modules/pkg.py
# Desc: manage BSD packages with pkg_*

'''
Manage BSD packages and repositories. Note that BSD package names are case-sensitive.
'''

from __future__ import unicode_literals

from pyinfra.api import operation

from .util.packaging import ensure_packages


@operation
def packages(state, host, packages=None, present=True, pkg_path=None):
    '''
    Install/remove/update pkg_* packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + pkg_path: the PKG_PATH environment variable to set

    pkg_path:
        By default this is autogenerated as follows (tested/working for OpenBSD):
        ``http://ftp.<OS>.org/pub/<OS>/<VERSION>/packages/<ARCH>/``. Note that OpenBSD's
        official mirrors only hold the latest two versions packages.

        NetBSD/FreeBSD helpfully use their own directory structures, so the default won't
        work.
    '''

    if present is True:
        # Autogenerate package path
        if not pkg_path:
            host_os = host.fact.os or ''
            pkg_path = 'http://ftp.{http}.org/pub/{os}/{version}/packages/{arch}/'.format(
                http=host_os.lower(),
                os=host_os,
                version=host.fact.os_version,
                arch=host.fact.arch,
            )

    yield ensure_packages(
        packages, host.fact.pkg_packages, present,
        install_command='PKG_PATH={0} pkg_add'.format(pkg_path),
        uninstall_command='pkg_delete',
    )



# pyinfra
# File: pyinfra/modules/gem.py
# Desc: manage gem packages

'''
Mange gem packages.
'''

from pyinfra.api import operation

from .util.packaging import ensure_packages


@operation
def packages(state, host, packages=None, present=True, latest=False):
    '''
    Add/remove/update gem packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version

    Versions:
        Package versions can be pinned like gem: ``<pkg>:<version>``.
    '''

    yield ensure_packages(
        packages, host.fact.gem_packages, present,
        install_command='gem install',
        uninstall_command='gem uninstall',
        upgrade_command='gem update',
        version_join=':',
        latest=latest,
    )



from __future__ import division

import math
import os
import sys

from collections import deque
from contextlib import contextmanager
from threading import Event, Thread
from time import sleep

import pyinfra

WAIT_TIME = 1 / 20
WAIT_CHARS = deque(('-', '/', '|', '\\'))

# Hacky way of getting terminal size (so can clear lines)
# Source: http://stackoverflow.com/questions/566746
TERMINAL_SIZE = os.popen('stty size', 'r').read().split()
IS_TTY = bool(TERMINAL_SIZE)
TERMINAL_WIDTH = (
    int(TERMINAL_SIZE[1])
    if IS_TTY else 0
)


def _print_spinner(stop_event, progress_queue):
    if not IS_TTY or os.environ.get('PYINFRA_PROGRESS') == 'off':
        return

    progress = ''
    text = ''

    while True:
        # Stop when asked too
        if stop_event.is_set():
            break

        WAIT_CHARS.rotate(1)

        try:
            progress = progress_queue[-1]
        except IndexError:
            pass

        text = '    {0}'.format(
            ' '.join((WAIT_CHARS[0], progress)),
        )
        text = '{0}\r'.format(text)

        sys.stdout.write(text)
        sys.stdout.flush()

        # In pyinfra_cli's __main__ we set stdout & stderr to be line buffered,
        # so write this escape code (clear line) into the buffer but don't flush,
        # such that any next print/log/etc clear the line first.
        sys.stdout.write('\033[K')
        sys.stderr.write('\033[K')

        sleep(WAIT_TIME)


@contextmanager
def progress_spinner(items, prefix_message=None):
    # If there's no pseudo state we're not in CLI mode, so just return a noop
    # handler and exit.
    if not pyinfra.is_cli:
        yield lambda complete_item: None
        return

    if not isinstance(items, set):
        items = set(items)

    total_items = len(items)
    stop_event = Event()

    def make_progress_message(include_items=True):
        message_bits = []

        # If we only have 1 item, don't show %
        if total_items > 1:
            percentage_complete = 0
            complete = total_items - len(items)
            percentage_complete = int(math.floor(complete / total_items * 100))
            message_bits.append('{0}% ({1}/{2})'.format(
                percentage_complete,
                complete,
                total_items,
            ))

        if prefix_message:
            message_bits.append(prefix_message)

        if include_items and items:
            # Plus 3 for the " - " joining below
            message_length = sum((len(message) + 3) for message in message_bits)
            # -8 for padding left+right, -2 for {} wrapping
            items_allowed_width = TERMINAL_WIDTH - 10 - message_length

            if items_allowed_width > 0:
                items_string = '{%s}' % (', '.join('{0}'.format(i) for i in items))
                if len(items_string) >= items_allowed_width:
                    items_string = '%s...}' % (
                        # -3 for the ...
                        items_string[:items_allowed_width - 3],
                    )

                message_bits.append(items_string)

        return ' - '.join(message_bits)

    progress_queue = deque((make_progress_message(),))

    def progress(complete_item):
        if complete_item not in items:
            raise ValueError('Invalid complete item: {0} not in {1}'.format(
                complete_item, items,
            ))

        items.remove(complete_item)
        progress_queue.append(make_progress_message())

    # Kick off the spinner thread
    spinner_thread = Thread(
        target=_print_spinner,
        args=(stop_event, progress_queue),
    )
    spinner_thread.daemon = True
    spinner_thread.start()

    # Yield allowing the actual code the spinner waits for to run
    yield progress

    # Finally, stop the spinner
    stop_event.set()
    spinner_thread.join()



import re

from configparser import ConfigParser

from pyinfra import logger


def _parse_ansible_hosts(hosts):
    for host in hosts:
        expand_match = re.search(r'\[[0-9:]+\]', host)
        if expand_match:
            expand_string = host[expand_match.start():expand_match.end()]
            bits = expand_string[1:-1].split(':')  # remove the [] either side

            zfill = 0
            if bits[0].startswith('0'):
                zfill = len(bits[0])

            start, end = int(bits[0]), int(bits[1])
            step = int(bits[2]) if len(bits) > 2 else 1

            for n in range(start, end + 1, step):
                number_as_string = '{0}'.format(n)
                if zfill:
                    number_as_string = number_as_string.zfill(zfill)

                hostname = host.replace(expand_string, number_as_string)
                yield hostname
            continue
        yield host


def load_ansible_ini_inventory(inventory_filename):
    logger.info('Trying Ansible inventory format...')

    config = ConfigParser(
        delimiters=(' '),  # we only handle the hostnames for now
        allow_no_value=True,  # we don't by default have = values
        interpolation=None,  # remove any Python interpolation
    )
    config.read(inventory_filename)

    groups = {}

    # First pass - load hosts/groups of hosts
    for section in config.sections():
        if ':' in section:  # ignore :children and :vars sections this time
            continue

        all_hosts = []
        options = config.options(section)
        for option in options:
            all_hosts.extend(_parse_ansible_hosts(options))

        groups[section] = all_hosts

    # Second pass - load any children groups
    for section in config.sections():
        if not section.endswith(':children'):  # we only support :children for now
            continue

        group_name = section.replace(':children', '')

        all_hosts = []
        options = config.options(section)
        for option in options:
            all_hosts.extend(groups[option])

        groups[group_name] = all_hosts

    return groups



from pyinfra.api import operation


@operation
def agent(state, host, server=None, port=None):
    '''
    Run puppet agent

    + server: master server URL
    + port: puppet master port
    '''

    args = []

    if server:
        args.append('--server=%s' % server)
    if port:
        args.append('--masterport=%s' % port)

    yield 'puppet agent -t %s' % ' '.join(args)



# pyinfra
# File: tests/test_cli.py
# Desc: tests for the pyinfra CLI

from collections import defaultdict
from datetime import datetime
from os import path
from random import shuffle
from unittest import TestCase

import six

from click.testing import CliRunner
from six.moves import cStringIO as StringIO

import pyinfra

from pyinfra import pseudo_state
from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all
from pyinfra.modules import server
from pyinfra_cli.exceptions import CliError
from pyinfra_cli.legacy import setup_arguments
from pyinfra_cli.main import cli
from pyinfra_cli.util import get_operation_and_args, json_encode, load_deploy_file

from .paramiko_util import PatchSSHTestCase
from .util import make_inventory


class TestCliDeployOperations(PatchSSHTestCase):
    def test_deploy(self):
        # Run 10 iterations of the test - each time shuffling the order of the
        # hosts - ensuring that the ordering has no effect on the operation order.
        for _ in range(10):
            self._do_test_deploy()

    def _do_test_deploy(self):
        correct_op_name_and_host_names = [
            ('First main operation', True),  # true for all hosts
            ('Second main operation', ('somehost',)),
            ('tests/test_deploy/a_task.py | First task operation', ('anotherhost',)),
            ('tests/test_deploy/a_task.py | Second task operation', ('anotherhost',)),
            ('tests/test_deploy/a_task.py | First task operation', True),
            ('tests/test_deploy/a_task.py | Second task operation', True),
            ('Loop-0 main operation', True),
            ('Loop-1 main operation', True),
            ('Third main operation', True),
        ]

        hosts = ['somehost', 'anotherhost', 'someotherhost']
        shuffle(hosts)

        inventory = make_inventory(hosts=hosts)
        state = State(inventory, Config())
        state.deploy_dir = path.join('tests', 'test_deploy')

        connect_all(state)
        pseudo_state.set(state)

        pyinfra.is_cli = True
        load_deploy_file(state, path.join(state.deploy_dir, 'deploy.py'))
        pyinfra.is_cli = False

        op_order = state.get_op_order()

        for i, (correct_op_name, correct_host_names) in enumerate(
            correct_op_name_and_host_names,
        ):
            op_hash = op_order[i]
            op_meta = state.op_meta[op_hash]

            self.assertEqual(list(op_meta['names'])[0], correct_op_name)

            for host in inventory:
                op_hashes = state.meta[host]['op_hashes']
                if correct_host_names is True or host.name in correct_host_names:
                    self.assertIn(op_hash, op_hashes)
                else:
                    self.assertNotIn(op_hash, op_hashes)


class TestCliUtil(TestCase):
    def test_json_encode_function(self):
        self.assertEqual(
            json_encode(setup_arguments),
            'setup_arguments',
        )

    def test_json_encode_datetime(self):
        now = datetime.utcnow()

        self.assertEqual(
            json_encode(now),
            now.isoformat(),
        )

    def test_json_encode_file(self):
        file = StringIO()

        self.assertEqual(
            json_encode(file),
            'In memory file: ',
        )

    def test_json_encode_set(self):
        self.assertEqual(
            json_encode({1, 2, 3}),
            [1, 2, 3],
        )

    def test_setup_no_module(self):
        with self.assertRaises(CliError) as context:
            get_operation_and_args(('no.op',))

        self.assertEqual(
            context.exception.message,
            'No such module: no',
        )

    def test_setup_no_op(self):
        with self.assertRaises(CliError) as context:
            get_operation_and_args(('server.no',))

        self.assertEqual(
            context.exception.message,
            'No such operation: server.no',
        )

    def test_setup_op_and_args(self):
        commands = ('server.user', 'one', 'two', 'hello=world')

        self.assertEqual(
            get_operation_and_args(commands),
            (
                server.user,
                (['one', 'two'], {'hello': 'world'}),
            ),
        )

    def test_setup_op_and_json_args(self):
        commands = ('server.user', '[["one", "two"], {"hello": "world"}]')

        self.assertEqual(
            get_operation_and_args(commands),
            (
                server.user,
                (['one', 'two'], {'hello': 'world'}),
            ),
        )


class TestCliExceptions(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_cli = CliRunner()
        cls.old_cli_show = CliError.show

    @classmethod
    def tearDownClass(cls):
        CliError.show = cls.old_cli_show

    def setUp(self):
        self.exception = None
        CliError.show = lambda e: self.capture_cli_error(e)

    def capture_cli_error(self, e):
        self.exception = e
        self.old_cli_show()

    def assert_cli_exception(self, args, message):
        self.test_cli.invoke(cli, args)

        self.assertIsInstance(self.exception, CliError)
        self.assertEqual(self.exception.message, message)

    def test_bad_deploy_file(self):
        self.assert_cli_exception(
            ['my-server.net', 'nop.py'],
            'No deploy file: nop.py',
        )

    def test_bad_fact(self):
        self.assert_cli_exception(
            ['my-server.net', 'fact', 'thing'],
            'No fact: thing',
        )


class TestLegacyCliArguments(TestCase):
    def make_cli_arguments(self, arguments):
        default_arguments = defaultdict(lambda: None)
        default_arguments.update(arguments)

        return default_arguments

    def assert_valid_arguments(self, arguments, check_arguments=None):
        arguments = self.make_cli_arguments(arguments)
        parsed_arguments = setup_arguments(arguments)

        if check_arguments:
            for key, value in six.iteritems(check_arguments):
                self.assertEqual(parsed_arguments[key], value)

    def test_int_arguments(self):
        self.assert_valid_arguments({
            '--parallel': '5',
            '--port': '22',
            '--fail-percent': '30',
        }, {
            'parallel': 5,
            'port': 22,
            'fail_percent': 30,
        })

    def test_wrong_int_argument(self):
        with self.assertRaises(
            CliError, message='string is not a valid integer for --parallel',
        ):
            self.assert_valid_arguments({
                '--parallel': 'string',
            })

    def test_invalid_files(self):
        with self.assertRaises(CliError, message='Deploy file not found: nop.py'):
            self.assert_valid_arguments({
                'DEPLOY': 'nop.py',
            })

        with self.assertRaises(
            CliError, message='Private key file not found: nop.key',
        ):
            self.assert_valid_arguments({
                '--key': 'nop.key',
            })

    def test_legacy_setup_op_and_args(self):
        op_string = 'server.user'
        args_string = 'one,two,hello=world'

        self.assertEqual(
            get_operation_and_args((op_string, args_string)),
            (
                server.user,
                (['one', 'two'], {'hello': 'world'}),
            ),
        )

    def test_legacy_setup_op_and_args_list(self):
        op_string = 'server.user'
        args_string = '[one,two],hello=world'

        self.assertEqual(
            get_operation_and_args((op_string, args_string)),
            (
                server.user,
                ([['one', 'two']], {'hello': 'world'}),
            ),
        )

    def test_legacy_setup_op_and_json_args(self):
        op_string = 'server.user'
        args_string = '[["one", "two"], {"hello": "world"}]'

        self.assertEqual(
            get_operation_and_args((op_string, args_string)),
            (
                server.user,
                (['one', 'two'], {'hello': 'world'}),
            ),
        )



# pyinfra
# File: pyinfra/api/__init__.py
# Desc: import some stuff

from .config import Config  # noqa: F401
from .deploy import deploy  # noqa: F401
from .exceptions import (  # noqa: F401
    DeployError,
    InventoryError,
    OperationError,
)
from .facts import FactBase, ShortFactBase  # noqa: F401
from .inventory import Inventory  # noqa: F401
from .operation import operation  # noqa: F401
from .state import State  # noqa: F401



# pyinfra
# File: pyinfra/local.py
# Desc: run stuff locally, within the context of operations - utility for the CLI

from __future__ import print_function, unicode_literals

from os import path
from subprocess import PIPE, Popen, STDOUT

import six

import pyinfra

from . import logger, pseudo_host, pseudo_state
from .api.exceptions import PyinfraError
from .api.util import ensure_host_list, get_caller_frameinfo, read_buffer


def include(filename, hosts=False, when=True):
    '''
    Executes a local python file within the ``pyinfra.pseudo_state.deploy_dir``
    directory.

    Args:
        hosts (string, list): group name or list of hosts to limit this include to
        when (bool): indicate whether to trigger operations in this include
    '''

    if not pyinfra.is_cli:
        raise PyinfraError('local.include is only available in CLI mode.')

    if not when:
        return

    if hosts is not False:
        hosts = ensure_host_list(hosts, inventory=pseudo_state.inventory)
        if pseudo_host not in hosts:
            return

    if pseudo_state.deploy_dir:
        filename = path.join(pseudo_state.deploy_dir, filename)

    frameinfo = get_caller_frameinfo()

    logger.debug('Including local file: {0}'.format(filename))

    try:
        # Fixes a circular import because `pyinfra.local` is really a CLI
        # only thing (so should be `pyinfra_cli.local`). It is kept here
        # to maintain backwards compatability and the nicer public import
        # (ideally users never need to import from `pyinfra_cli`).

        from pyinfra_cli.config import extract_file_config
        from pyinfra_cli.util import exec_file

        # Load any config defined in the file and setup like a @deploy
        config_data = extract_file_config(filename)
        kwargs = {
            key.lower(): value
            for key, value in six.iteritems(config_data)
            if key in [
                'SUDO', 'SUDO_USER', 'SU_USER',
                'PRESERVE_SUDO_ENV', 'IGNORE_ERRORS',
            ]
        }
        with pseudo_state.deploy(
            filename, kwargs, None, frameinfo.lineno,
            in_deploy=False,
        ):
            exec_file(filename)

        # One potential solution to the above is to add local as an actual
        # module, ie `pyinfra.modules.local`.

    except IOError as e:
        raise PyinfraError(
            'Could not include local file: {0}\n{1}'.format(filename, e),
        )


def shell(commands, splitlines=False, ignore_errors=False):
    '''
    Subprocess based implementation of pyinfra/api/ssh.py's ``run_shell_command``.

    Args:
        commands (string, list): command or list of commands to execute
        spltlines (bool): optionally have the output split by lines
        ignore_errors (bool): ignore errors when executing these commands
    '''

    if isinstance(commands, six.string_types):
        commands = [commands]

    all_stdout = []

    # Checking for pseudo_state means this function works outside a deploy
    # eg the vagrant connector.
    print_output = (
        pseudo_state.print_output
        if pseudo_state.isset()
        else False
    )

    for command in commands:
        print_prefix = 'localhost: '

        if print_output:
            print('{0}>>> {1}'.format(print_prefix, command))

        process = Popen(command, shell=True, stdout=PIPE, stderr=STDOUT)

        stdout = read_buffer(
            process.stdout,
            print_output=print_output,
            print_func=lambda line: '{0}{1}'.format(print_prefix, line),
        )

        # Get & check result
        result = process.wait()

        # Close any open file descriptor
        process.stdout.close()

        if result > 0 and not ignore_errors:
            raise PyinfraError(
                'Local command failed: {0}\n{1}'.format(command, stdout),
            )

        all_stdout.extend(stdout)

    if not splitlines:
        return '\n'.join(all_stdout)

    return all_stdout



# pyinfra
# File: pyinfra/facts/gem.py
# Desc: facts for the RubyGems package manager

from pyinfra.api import FactBase

from .util.packaging import parse_packages

GEM_REGEX = r'^([a-zA-Z0-9\-\+\_]+)\s\(([0-9\.]+)\)$'


class GemPackages(FactBase):
    '''
    Returns a dict of installed gem packages:

    .. code:: python

        'package_name': 'version',
        ...
    '''

    command = 'gem list --local'
    default = dict

    def process(self, output):
        return parse_packages(GEM_REGEX, output)
